/* serverpacket.c
 *
 * Constuct and send DHCP server packets
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>

//#include <net/ipelios/ipelios.h>
#include <net/dhcpelios/debug.h>
#include "packet.h"
#include "dhcpd.h"
#include "options.h"
#include "leases.h"
#include "ecos.h"

//------------------------------------------------------------------------------------------------------
// send a packet to giaddr using the kernel ip stack
//------------------------------------------------------------------------------------------------------
static int send_packet_to_relay(struct dhcpMessage *payload)
{
#ifdef LOGACTIVE
 char gw[IPSTRING];
#endif

	LOG(LOG_INFO,"send_packet_to_relay","Called");
	LOG(LOG_INFO,"send_packet_to_relay","Forwarding packet to relay :%s",Number2String(payload->giaddr,gw));

	return kernel_packet(payload, server_config.server, SERVER_PORT,payload->giaddr, SERVER_PORT);
}

//------------------------------------------------------------------------------------------------------
// send a packet to a specific arp address and ip address by creating our own ip packet
// force_broadcast is a boolean to activate broadcast
//------------------------------------------------------------------------------------------------------
static int send_packet_to_client(struct dhcpMessage *payload, int force_broadcast)
{
	unsigned char *chaddr=NULL;	// MAC target
	u_int32_t ciaddr=0;		// IP target
#ifdef LOGACTIVE
	char ip[IPSTRING];		// IP adress conversion
#endif

	LOG(LOG_INFO,"send_packet_to_client","Called with force_broadcast:%d",force_broadcast);
	if (force_broadcast)
	{
		LOG(LOG_INFO,"send_packet_to_client","Broadcasting packet to client (NAK)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else if (payload->ciaddr)
	{
		chaddr = payload->chaddr;
		ciaddr = payload->ciaddr;
		LOG(LOG_INFO,"send_packet_to_client","Unicasting packet to client ciaddr '%s' [%02x:%02x:%02x:%02x:%02x:%02x]",Number2String(ciaddr,ip),chaddr[0],chaddr[1],chaddr[2],chaddr[3],chaddr[4],chaddr[5]);

	} else if (ntohs(payload->flags) & BROADCAST_FLAG)
	{
		LOG(LOG_INFO,"send_packet_to_client","Broadcasting packet to client (requested)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else
	{
		ciaddr = payload->yiaddr;
		chaddr = payload->chaddr;
		LOG(LOG_INFO,"send_packet_to_client","Unicasting packet to client yiaddr '%s' [%02x:%02x:%02x:%02x:%02x:%02x]",Number2String(ciaddr,ip),chaddr[0],chaddr[1],chaddr[2],chaddr[3],chaddr[4],chaddr[5]);
	}
	return raw_packet(payload, server_config.server, SERVER_PORT,ciaddr, CLIENT_PORT, chaddr, server_config.ifindex);
}

//------------------------------------------------------------------------------------------------------
// send a dhcp packet, if force broadcast is set, the packet will be broadcasted to the client
//------------------------------------------------------------------------------------------------------
static int send_packet(struct dhcpMessage *payload, int force_broadcast)
{
	int ret=0;
	LOG(LOG_INFO,"send_packet","Called with force_broadcast :%d",force_broadcast);

	if (payload->giaddr)
		ret = send_packet_to_relay(payload);
	else ret = send_packet_to_client(payload, force_broadcast);
	return (ret);
}

//------------------------------------------------------------------------------------------------------
// Init a DHCP packet with an old one
//------------------------------------------------------------------------------------------------------
static void init_packet(struct dhcpMessage *packet, struct dhcpMessage *oldpacket, char type)
{
	LOG(LOG_INFO,"init_packet","Called with type :%02x",type);

	init_header(packet, type);
	packet->xid = oldpacket->xid;
	LOG(LOG_INFO,"init_packet","ID request :%08x",packet->xid);
	memcpy(packet->chaddr, oldpacket->chaddr, 16);
	packet->flags = oldpacket->flags;
	packet->giaddr = oldpacket->giaddr;
	packet->ciaddr = oldpacket->ciaddr;
	add_simple_option(packet->options, DHCP_SERVER_ID, server_config.server);
}

//------------------------------------------------------------------------------------------------------
// Add in the bootp options
//------------------------------------------------------------------------------------------------------
static void add_bootp_options(struct dhcpMessage *packet)
{
	LOG(LOG_INFO, "add_bootp_options","Called");

	packet->siaddr = server_config.siaddr;
	if (server_config.sname)
		strncpy(packet->sname, server_config.sname, sizeof(packet->sname) - 1);
	if (server_config.boot_file)
		strncpy(packet->file, server_config.boot_file, sizeof(packet->file) - 1);
}

//------------------------------------------------------------------------------------------------------
// send a DHCP OFFER to a DHCP DISCOVER
//------------------------------------------------------------------------------------------------------
int sendOffer(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;			//  New packet
	struct dhcpOfferedAddr *lease = NULL;
	u_int32_t req_align=0;
	u_int32_t lease_time_align = server_config.lease;
	unsigned char *req=NULL;
	unsigned char *lease_time=NULL;
	struct option_set *curr=NULL;
	struct in_addr addr;
#ifdef LOGACTIVE
	char ip[IPSTRING];		// IP adress conversion
#endif


	LOG(LOG_INFO, "sendOffer","Called");

	init_packet(&packet, oldpacket, DHCPOFFER);

	// ADDME: if static, short circuit
	// the client is in our lease/offered table
	if ((lease = find_lease_by_chaddr(oldpacket->chaddr)))
        {
		if (!lease_expired(lease))
			lease_time_align = lease->expires - time(0);
		packet.yiaddr = lease->yiaddr;

	// Or the client has a requested ip
	} else if ((req = get_option(oldpacket, DHCP_REQUESTED_IP)) &&

		   // Don't look here (ugly hackish thing to do)
		   memcpy(&req_align, req, 4) &&

		   // and the ip is in the lease range
		   ntohl(req_align) >= ntohl(server_config.start) &&
		   ntohl(req_align) <= ntohl(server_config.end) &&

		   // and its not already taken/offered
		   // ADDME: check that its not a static lease
		   ((!(lease = find_lease_by_yiaddr(req_align)) ||

		   // or its taken, but expired
		   // ADDME: or maybe in here
		   lease_expired(lease))))
	{
	 packet.yiaddr = req_align; // FIXME: oh my, is there a host using this IP?

	// otherwise, find a free IP
	// ADDME: is it a static lease?
	}
	else
	{
		LOG(LOG_INFO,"sendOffer","Seeking an adress");
		packet.yiaddr = find_address(0);

		// try for an expired lease
		if (!packet.yiaddr) packet.yiaddr = find_address(1);
		LOG(LOG_INFO,"sendOffer","Find adress '%s'",Number2String(packet.yiaddr,ip));
	}
	if(!packet.yiaddr)
	{
		LOG(LOG_WARNING,"sendOffer","No IP addresses to give -- OFFER abandoned");
		return (-1);
	}

	if (!add_lease(packet.chaddr, packet.yiaddr, server_config.offer_time,16))
	{
		LOG(LOG_WARNING,"sendOffer","Lease pool is full -- OFFER abandoned");
		return (-1);
	}

	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME)))
	 {
		memcpy(&lease_time_align, lease_time, 4);
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config.lease)
			lease_time_align = server_config.lease;
	}
	LOG(LOG_INFO,"sendOffer","Waking an address");
	// Make sure we aren't just using the lease time from the previous offer
	if (lease_time_align < server_config.min_lease)
		lease_time_align = server_config.lease;
	// ADDME: end of short circuit
	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));
	LOG(LOG_INFO,"sendOffer","Simple Option added");

	// Seek LEASE TIME in options
	curr = server_config.options;
	while (curr)
	{
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet);
	LOG(LOG_INFO,"sendOffer","BOOT Option added");

	addr.s_addr = packet.yiaddr;
	LOG(LOG_INFO,"sendOffer","Sending OFFER of %s", Number2String(packet.yiaddr,ip));
	return send_packet(&packet, 0);
}

//------------------------------------------------------------------------------------------------------
// Sending a NAK message
//------------------------------------------------------------------------------------------------------
int sendNAK(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	LOG(LOG_INFO,"sendNAK","Called");
	init_packet(&packet, oldpacket, DHCPNAK);

	return send_packet(&packet, 1);
}

//------------------------------------------------------------------------------------------------------
// Sending a ACK message
//------------------------------------------------------------------------------------------------------
int sendACK(struct dhcpMessage *oldpacket, u_int32_t yiaddr)
{
	struct dhcpMessage packet;
	struct option_set *curr=NULL;
	unsigned char *lease_time=NULL;
	u_int32_t lease_time_align = server_config.lease;
	struct in_addr addr;

	init_packet(&packet, oldpacket, DHCPACK);
	packet.yiaddr = yiaddr;

	LOG(LOG_INFO, "sendACK","Called");
	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME)))
	{
		memcpy(&lease_time_align, lease_time, 4);
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config.lease)
			lease_time_align = server_config.lease;
		else if (lease_time_align < server_config.min_lease)
			lease_time_align = server_config.lease;
	}

	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	// Seek LEASE TIME in options
	curr = server_config.options;
	while (curr)
	{
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet);

	addr.s_addr = packet.yiaddr;
	LOG(LOG_INFO,"sendACK","Sending ACK to %s", inet_ntoa(addr));

	if (send_packet(&packet, 0) < 0)
		return (-1);

	add_lease(packet.chaddr, packet.yiaddr, lease_time_align,16);

	return (0);
}

//------------------------------------------------------------------------------------------------------
// Sending information message
//------------------------------------------------------------------------------------------------------
int send_inform(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	struct option_set *curr=NULL;

	LOG(LOG_INFO, "send_inform","Called");
	init_packet(&packet, oldpacket, DHCPACK);

	// Seek LEASE TIME in options
	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet);

	return send_packet(&packet, 0);
}

