/*
 * socket.c -- DHCP server client/server socket creation
 *
 * udhcp client/server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 * Ported to eCOS by Sebastien Couret <sebastien.couret@elios-informatique.fr> November 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <network.h>
#include <net/ethernet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <errno.h>
#include "ecos.h"
//#include <net/ipelios/ipelios.h>
#include <net/dhcpelios/debug.h>

//-----------------------------------------------------------------------------------------
// Read index,IP address and MAC adress of interface
// Return 0 on success, -1 if failed
//-----------------------------------------------------------------------------------------
int read_interface(char *interface, int *ifindex, u_int32_t *addr, unsigned char *arp)
{
	int fd=0;
	struct ifreq ifr;
	struct sockaddr_in *our_ip=NULL;

	LOG(LOG_INFO,"read_interface","Called with interface :'%s'",interface);
	memset(&ifr, 0, sizeof(struct ifreq));

	if((fd = socket(AF_INET, SOCK_DGRAM, 0)) >= 0)
	{
		ifr.ifr_addr.sa_family = AF_INET;
		strncpy(ifr.ifr_name, interface,IFNAMSIZ);

		if (addr)
		{
			// Get interface address
			if (ioctl(fd, SIOCGIFADDR, &ifr) == 0)
			{
				our_ip = (struct sockaddr_in *) &ifr.ifr_addr;
				*addr = our_ip->sin_addr.s_addr;
				LOG(LOG_INFO, "read_interface","Our IP on '%s' is '%s'", ifr.ifr_name, inet_ntoa(our_ip->sin_addr));
			}
			else
			{
				LOG(LOG_ERR, "read_interface"," SIOCGIFADDR failed, is the interface '%s' up and configured?: '%s'",interface,strerror(errno));
				close(fd);
				return (-1);
			}
		}

		*ifindex = if_nametoindex(interface);
		LOG(LOG_INFO,"read_interface", "Adapter index is %d",*ifindex);

		// Get Hardware adress
		if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0)
		{
		 memcpy(arp, ifr.ifr_hwaddr.sa_data,ETHER_ADDR_LEN);
		 LOG(LOG_INFO,"read_interface", "Adapter hardware address '%02x:%02x:%02x:%02x:%02x:%02x'",arp[0], arp[1], arp[2], arp[3], arp[4], arp[5]);
		}
		else
		{
		 LOG(LOG_ERR,"read_interface", "SIOCGIFHWADDR failed!: '%s'", strerror(errno));
	 	 close(fd);
		 return (-1);
		}
  	 close(fd);
	} else
	{
	 LOG(LOG_ERR, "read_interface","Getting interface informations failed!: '%s'", strerror(errno));
	 return (-1);
	}
	return (0);
}

//-----------------------------------------------------------------------------------------
// Create a listening UDP broadcast socket on selected ip port [and interface]
//-----------------------------------------------------------------------------------------
int listen_socket(unsigned int ip, int port, char *inf)
{
	int fd=0;
	struct sockaddr_in addr;
	int n = 1;
#ifdef LOGACTIVE
	char ips[IPSTRING];
#endif

	LOG(LOG_INFO, "listen_socket","Called with '%s':%d inf='%s'",  Number2String(ip,ips), port, inf);
	if ((fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	{
	 LOG(LOG_ERR,"listen_socket", "Socket call failed: '%s'", strerror(errno));
	 return (-1);
	}
	LOG(LOG_INFO, "listen_socket","Listen socket is %d", fd);

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof(n)) == -1)
	{
	 close(fd);
	 LOG(LOG_ERR,"listen_socket", "Setsockopt for REUSEADDR failed: '%s'",strerror(errno));
	 return (-1);
	}
 	LOG(LOG_INFO,"listen_socket", "Socket will work for BROADCAST");
	if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char *) &n, sizeof(n)) == -1)
	{
	 close(fd);
	 LOG(LOG_ERR,"listen_socket", "Setsockopt for BROADCAST failed :'%s'",strerror(errno));
	 return (-1);
	}

	// Adresse Source
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = ip;
	if (bind(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr)) == -1)
	{
		LOG(LOG_ERR,"listen_socket", "Bind failed :'%s'",strerror(errno));
		close(fd);
		return (-1);
	}
	return (fd);
}


//-----------------------------------------------------------------------------------------
// Open a raw socket
//-----------------------------------------------------------------------------------------
int raw_socket(int ifindex)
{
 int fd=0;
 struct sockaddr_dl sock;

 LOG(LOG_INFO,"raw_socket","Called with ifindex %d\n", ifindex);
 // 3eme arg=htons(ETH_P_IP)
 if ((fd = socket(AF_INET, SOCK_RAW,htons(ETHERTYPE_IP))) < 0) {
  LOG(LOG_ERR, "raw_socket","Socket call failed: '%s'", strerror(errno));
  return (-1);
 }

 sock.sdl_family=AF_INET;
// sock.sdl_type=ETHERTYPE_IP; // SCT: Warning type there !
 sock.sdl_index = ifindex;
 sock.sdl_len=sizeof(sock);
 if (bind(fd, (struct sockaddr *) &sock, sizeof(sock)) < 0)
 {
  LOG(LOG_ERR,"raw_socket","Bind call failed: '%s'", strerror(errno));
  close(fd);
  return (-1);
 }
 LOG(LOG_INFO,"raw_socket","Raw socket opened :%d",fd);
 return (fd);
}

